import typing

import numpy as np
import pandas as pd

from ._alifestd_categorize_triplet_asexual import (
    alifestd_categorize_triplet_asexual,
)
from ._alifestd_find_leaf_ids import alifestd_find_leaf_ids
from ._estimate_binomial_p import estimate_binomial_p


def alifestd_estimate_triplet_distance_asexual(
    first_df: pd.DataFrame,
    second_df: pd.DataFrame,
    taxon_label_key: str,
    confidence: float = 0.99,
    precision: float = 0.01,
    strict: bool = True,
    detail: bool = False,
    progress_wrap: typing.Callable = lambda x: x,
    mutate: bool = False,
) -> typing.Union[typing.Tuple[float, typing.Tuple[float, float, int]], float]:
    if not mutate:
        first_df = first_df.copy()
        second_df = second_df.copy()

    first_df = first_df.reset_index(drop=True).set_index("id", drop=False)
    second_df = second_df.reset_index(drop=True).set_index("id", drop=False)

    first_leaf_ids = alifestd_find_leaf_ids(first_df)
    first_taxon_labels = first_df.loc[first_leaf_ids, taxon_label_key]
    second_leaf_ids = alifestd_find_leaf_ids(second_df)
    second_taxon_labels = second_df.loc[second_leaf_ids, taxon_label_key]

    assert first_taxon_labels.is_unique
    assert second_taxon_labels.is_unique
    assert (
        first_taxon_labels.sort_values().to_numpy()
        == second_taxon_labels.sort_values().to_numpy()
    ).all()

    if len(first_leaf_ids) < 3:
        return (0.0, [0.0, 0.0], 0) if detail else 0.0

    def sample_triplet_comparison() -> bool:
        labels = np.random.choice(first_taxon_labels, 3, replace=False)
        first_triplet_ids = [
            first_df.index[first_df[taxon_label_key] == label].item()
            for label in labels
        ]
        second_triplet_ids = [
            second_df.index[second_df[taxon_label_key] == label].item()
            for label in labels
        ]
        assert 3 == len(first_triplet_ids) == len(second_triplet_ids)

        cat1 = alifestd_categorize_triplet_asexual(
            first_df, first_triplet_ids, mutate=True
        )
        cat2 = alifestd_categorize_triplet_asexual(
            second_df, second_triplet_ids, mutate=True
        )
        match = cat1 == cat2 or (not strict and -1 in (cat1, cat2))
        return not match

    res = estimate_binomial_p(
        sample_triplet_comparison,
        confidence=confidence,
        precision=precision,
        progress_wrap=progress_wrap,
    )
    return res if detail else res[0]



import numpy as np
import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_estimate_triplet_distance_asexual,
    alifestd_make_ancestor_list_col,
    alifestd_make_empty,
)


def test_empty():
    assert 0.0 == alifestd_estimate_triplet_distance_asexual(
        alifestd_make_empty(), alifestd_make_empty(), "id"
    )


def test_undersize():
    df = pd.DataFrame(
        {
            "id": [0, 1, 2],
            "taxon_label": [4, 5, 6],
            "ancestor_id": [0, 0, 0],
        },
    )
    df["ancestor_list"] = alifestd_make_ancestor_list_col(
        df["id"], df["ancestor_id"]
    )
    for xdf in df[:1], df[:2], df:
        assert (
            alifestd_estimate_triplet_distance_asexual(
                xdf, xdf.sample(frac=1), "taxon_label", detail=False
            )
            == 0.0
        )


@pytest.mark.parametrize(
    "df",
    [
        pd.DataFrame(
            {
                "id": [1, 2, 3],
                "taxon_label": ["a", "b", "c"],
                "ancestor_id": [1, 1, 1],
            },
        ),
        pd.DataFrame(
            {
                "id": [0, 1, 2, 3],
                "taxon_label": [0, 1, 2, 3],
                "ancestor_id": [0, 0, 0, 0],
            },
        ),
        pd.DataFrame(
            {
                "id": [0, 1, 2, 3, 4, 5],
                "taxon_label": ["0", "1", "2", "3", "4", "5"],
                "ancestor_id": [0, 0, 0, 0, 1, 2],
            },
        ),
        pd.DataFrame(
            {
                "id": [9, 1, 2, 3, 4, 5],
                "taxon_label": ["9", "1", "2", "3", "4", "5"],
                "ancestor_id": [9, 9, 9, 9, 1, 2],
            },
        ),
        pd.DataFrame(
            {
                "id": [0, 1, 2, 3, 4, 5, 6],
                "taxon_label": ["0", "1", "2", "3", "4", "5", "6"],
                "ancestor_id": [0, 0, 0, 0, 1, 2, 3],
            },
        ),
    ],
)
def test_polytomy_identical(df: pd.DataFrame):
    df["ancestor_list"] = alifestd_make_ancestor_list_col(
        df["id"], df["ancestor_id"]
    )
    for adf, bdf in (df, df), (df, df.sample(frac=1)):
        est, ci, n = alifestd_estimate_triplet_distance_asexual(
            adf, bdf, "taxon_label", confidence=0.95, detail=True
        )
        assert est == 0
        assert np.isclose(ci[0], 0)
        assert ci[1] >= ci[0]


@pytest.mark.parametrize("strict", [True, False])
def test_differing_wrong1(strict: bool):
    adf = pd.DataFrame(
        {
            "id": reversed([9, 1, 2, 3, 4, 5]),
            "taxon_label": reversed(["0", "1", "2", "3", "4", "5"]),
            "ancestor_id": reversed([9, 9, 1, 2, 2, 1]),
        },
    )
    adf["ancestor_list"] = alifestd_make_ancestor_list_col(
        adf["id"], adf["ancestor_id"]
    )
    bdf = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 5, 4],
            "taxon_label": ["0", "1", "2", "3", "5", "4"],
            "ancestor_id": [0, 0, 1, 2, 2, 1],
        },
    )
    bdf["ancestor_list"] = alifestd_make_ancestor_list_col(
        bdf["id"], bdf["ancestor_id"]
    )
    est = alifestd_estimate_triplet_distance_asexual(
        adf, bdf, "taxon_label", confidence=0.95, strict=strict
    )
    assert 1 == est

    est = alifestd_estimate_triplet_distance_asexual(
        adf, bdf.sample(frac=1), "taxon_label", confidence=0.95, strict=strict
    )
    assert 1 == est


@pytest.mark.parametrize("strict", [False])
def test_differing_wrong2(strict: bool):
    adf = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5],
            "taxon_label": ["0", "1", "2", "3", "4", "5"],
            "ancestor_id": [0, 0, 1, 2, 2, 1],
        },
    )
    adf["ancestor_list"] = alifestd_make_ancestor_list_col(
        adf["id"], adf["ancestor_id"]
    )
    bdf = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5],
            "taxon_label": ["0", "1", "2", "3", "4", "5"],
            "ancestor_id": [0, 0, 1, 2, 1, 2],
        },
    )
    bdf["ancestor_list"] = alifestd_make_ancestor_list_col(
        bdf["id"], bdf["ancestor_id"]
    )
    est = alifestd_estimate_triplet_distance_asexual(
        adf, bdf, "taxon_label", confidence=0.95, strict=strict
    )
    assert 1 == est

    est = alifestd_estimate_triplet_distance_asexual(
        adf, bdf.sample(frac=1), "taxon_label", confidence=0.95, strict=strict
    )
    assert 1 == est

    est = alifestd_estimate_triplet_distance_asexual(
        adf.sample(frac=1),
        bdf.sample(frac=1),
        "taxon_label",
        confidence=0.95,
        strict=strict,
    )
    assert 1 == est


@pytest.mark.parametrize("strict", [True, False])
def test_differing_polytomy(strict: bool):
    adf = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5],
            "taxon_label": ["0", "1", "2", "3", "4", "5"],
            "ancestor_id": [0, 0, 1, 2, 2, 1],
        },
    )
    adf["ancestor_list"] = alifestd_make_ancestor_list_col(
        adf["id"], adf["ancestor_id"]
    )
    bdf = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5],
            "taxon_label": ["0", "1", "2", "3", "4", "5"],
            "ancestor_id": [0, 0, 1, 2, 2, 2],
        },
    )
    bdf["ancestor_list"] = alifestd_make_ancestor_list_col(
        bdf["id"], bdf["ancestor_id"]
    )
    est = alifestd_estimate_triplet_distance_asexual(
        adf, bdf, "id", confidence=0.95, precision=0.05, strict=strict
    )
    assert bool(strict) == est

    est = alifestd_estimate_triplet_distance_asexual(
        adf,
        bdf.sample(frac=1),
        "taxon_label",
        confidence=0.95,
        precision=0.05,
        strict=strict,
    )
    assert bool(strict) == est


@pytest.mark.parametrize("strict", [True, False])
def test_differing_wrong_big(strict: bool):
    adf = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            "taxon_label": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            "ancestor_id": [0, 0, 1, 2, 2, 1, 5, 5, 4, 4],
        },
    )
    adf["ancestor_list"] = alifestd_make_ancestor_list_col(
        adf["id"], adf["ancestor_id"]
    )
    bdf = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            "taxon_label": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            "ancestor_id": [0, 0, 1, 5, 2, 1, 2, 5, 4, 4],
        },
    )
    bdf["ancestor_list"] = alifestd_make_ancestor_list_col(
        bdf["id"], bdf["ancestor_id"]
    )
    est = alifestd_estimate_triplet_distance_asexual(
        adf,
        bdf,
        "taxon_label",
        confidence=0.95,
        precision=0.05,
        strict=strict,
    )
    assert 0 < est < 1

    est = alifestd_estimate_triplet_distance_asexual(
        adf,
        bdf.sample(frac=1),
        "taxon_label",
        confidence=0.95,
        precision=0.05,
        strict=strict,
    )
    assert 0 < est < 1



import pickle
import random
import tempfile

import pytest

from hstrat.hstrat import stochastic_algo


@pytest.mark.parametrize(
    "replicate",
    range(5),
)
def test_eq(replicate):
    random.seed(replicate)
    spec = stochastic_algo.PolicySpec()
    assert spec == spec
    assert spec == stochastic_algo.PolicySpec()


def test_GetEvalCtor():
    # hstrat. is needed for eval()
    from hstrat import hstrat  # noqa

    spec = stochastic_algo.PolicySpec()
    reconstituted = eval(spec.GetEvalCtor())  # noqa
    eval_ctor = spec.GetEvalCtor()
    assert eval_ctor.startswith("hstrat.stochastic_algo.PolicySpec(")
    assert eval_ctor.endswith(")")
    reconstituted = eval(eval_ctor)  # noqa
    assert spec == reconstituted


def test_pickle():
    original = stochastic_algo.PolicySpec()
    with tempfile.TemporaryDirectory() as tmp_path:
        with open(f"{tmp_path}/data", "wb") as tmp_file:
            pickle.dump(original, tmp_file)

        with open(f"{tmp_path}/data", "rb") as tmp_file:
            reconstituted = pickle.load(tmp_file)
            assert reconstituted == original


@pytest.mark.parametrize(
    "replicate",
    range(5),
)
def test_init(replicate):
    random.seed(replicate)
    stochastic_algo.PolicySpec()


def test_GetAlgoIdentifier():
    spec = stochastic_algo.PolicySpec()
    assert spec.GetAlgoIdentifier()


def test_GetAlgoTitle():
    spec = stochastic_algo.PolicySpec()
    assert spec.GetAlgoTitle()


def test_repr():
    spec = stochastic_algo.PolicySpec()
    assert spec.GetAlgoIdentifier() in repr(spec)


def test_str():
    spec = stochastic_algo.PolicySpec()
    assert spec.GetAlgoTitle() in str(spec)
